----------------------------------------------------------------------------------------------------
-- DynamicCultivatorMode
----------------------------------------------------------------------------------------------------
-- Purpose: Specialization for dynamic cultivator mode setting by animation
--
-- @author John Deere 6930 @VertexDezign
----------------------------------------------------------------------------------------------------

---@class AttachableCylinderedParts
AttachableCylinderedParts = {}
AttachableCylinderedParts.MOD_NAME = g_currentModName

function AttachableCylinderedParts.initSpecialization(specializations)
    local schema = Vehicle.xmlSchema
    schema:setXMLSpecializationType("Cylindered")

    local partKey = "vehicle.cylindered.attachableCylinderedParts.part(?)"

    schema:register(XMLValueType.NODE_INDEX, partKey .. "#node", "Part node.")
    schema:register(XMLValueType.NODE_INDEX, partKey .. "#referenceFrame", "Part reference frame.")
    schema:register(XMLValueType.NODE_INDEX, partKey .. "#referencePoint", "Part reference point.")
    schema:register(XMLValueType.STRING, partKey .. "#referenceIdentName", "Part reference indentifier name.")
    
    schema:register(XMLValueType.VECTOR_N, partKey .. "#attacherJointIndices", "Attacher joint indicies.")
    schema:register(XMLValueType.VECTOR_N, partKey .. "#inputAttacherJointIndices", "Input attacher joint indicies.")
    

    schema:setXMLSpecializationType()
end

function AttachableCylinderedParts.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Cylindered, specializations)
end

function AttachableCylinderedParts.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "loadAttachablePart", AttachableCylinderedParts.loadAttachablePart)
	SpecializationUtil.registerFunction(vehicleType, "arePartsValid", AttachableCylinderedParts.arePartsValid)
	SpecializationUtil.registerFunction(vehicleType, "connectMovingParts", AttachableCylinderedParts.connectMovingParts)
	SpecializationUtil.registerFunction(vehicleType, "disconnectMovingParts", AttachableCylinderedParts.disconnectMovingParts)
end

function AttachableCylinderedParts.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", AttachableCylinderedParts)
    SpecializationUtil.registerEventListener(vehicleType, "onPostAttachImplement", AttachableCylinderedParts)
    SpecializationUtil.registerEventListener(vehicleType, "onPreDetachImplement", AttachableCylinderedParts)
end

---Called on load
function AttachableCylinderedParts:onLoad(savegame)
    self.spec_attachableCylinderedParts = self[string.format("spec_%s.attachableCylinderedParts", AttachableCylinderedParts.MOD_NAME)]
    local spec = self.spec_attachableCylinderedParts

    spec.attachableCylinderedParts = {}

    self.xmlFile:iterate("vehicle.cylindered.attachableCylinderedParts.part", function (_, partKey)
        local entry = {}
        if self:loadAttachablePart(self.xmlFile, partKey, entry) then
            table.insert(spec.attachableCylinderedParts, entry)
        end
    end)
end

---Loads attachable cylindered part from xml
---@param xmlFile table
---@param key any
---@param entry any
---@return boolean
function AttachableCylinderedParts:loadAttachablePart(xmlFile, key, entry)
    local node = xmlFile:getValue(key .. "#node", nil, self.components, self.i3dMappings)
    if node ~= nil then
        entry.node = node
    else
        local nodeValid = false
        local referenceFrame = xmlFile:getValue(key .. "#referenceFrame", nil, self.components, self.i3dMappings)

        if referenceFrame ~= nil then
            entry.referenceFrame = referenceFrame
            nodeValid = true
        end
        local referencePoint = xmlFile:getValue(key .. "#referencePoint", nil, self.components, self.i3dMappings)

        if referencePoint ~= nil then
            entry.referencePoint = referencePoint
            nodeValid = true
        end
        if not nodeValid then
            Logging.xmlWarning(xmlFile, "AttachableCylinderedPart '%s' no valid reference given!", key)
            return false
        end
    end

    entry.connected = false
    entry.connectedPart = nil

    local referenceIdentName = xmlFile:getValue(key .. "#referenceIdentName")
    if referenceIdentName ~= nil and referenceIdentName ~= "" then
        entry.referenceIdentName = referenceIdentName
    else
        Logging.xmlWarning(xmlFile, "AttachableCylinderedPart '%s' no referenceIdentName given!", key)
        return false
    end

    --load AttacherJoints
    entry.attacherJointIndices = {}
    local indices = xmlFile:getValue(key .. "#attacherJointIndices", nil, true)
    for _, index in ipairs(indices) do
        if self:getAttacherJointByJointDescIndex(index) ~= nil then
            entry.attacherJointIndices[index] = true
        end
    end

    --load inputAttacherJoints
    entry.inputAttacherJointIndices = {}
    indices = xmlFile:getValue(key .. "#inputAttacherJointIndices", nil, true)
    for _, index in ipairs(indices) do
        if self:getInputAttacherJointByJointDescIndex(index) ~= nil then
            entry.inputAttacherJointIndices[index] = true
        end
    end

    if #entry.inputAttacherJointIndices == 0 and #entry.attacherJointIndices == 0 then
        Logging.xmlWarning(xmlFile, "AttachableCylinderedPart '%s' at least one inputAttacherJoint or attacherJoint must be given!", key)
        return false
    end

    return true
end

---Called after object was attached
---@param attachableObject table
---@param inputJointDescIndex integer
---@param jointDescIndex integer
function AttachableCylinderedParts:onPostAttachImplement(attachableObject, inputJointDescIndex, jointDescIndex)
    local spec = self.spec_attachableCylinderedParts

    if self:arePartsValid(attachableObject) then
        local attachedParts = attachableObject.spec_attachableCylinderedParts

        for _, part in ipairs(spec.attachableCylinderedParts) do
            if not part.connected and part.attacherJointIndices[jointDescIndex] then
                for _, partAttached in ipairs(attachedParts.attachableCylinderedParts) do
                    if partAttached.inputAttacherJointIndices[inputJointDescIndex] then
                        if part.referenceIdentName == partAttached.referenceIdentName then
                            part.connected = self:connectMovingParts(part, partAttached, attachableObject)
                        end
                    end
                end
            end
        end
    end
end

---Called before detaching given implement
---@param implement table
function AttachableCylinderedParts:onPreDetachImplement(implement)
    local spec = self.spec_attachableCylinderedParts

    if self:arePartsValid(implement.object) then
        for _, part in ipairs(spec.attachableCylinderedParts) do
            if part.connected then
                part.connected = not self:disconnectMovingParts(part, implement.object)
            end
        end
    end
end

---Returns true if given object has valid parts
---@param attachableObject table
---@return boolean
function AttachableCylinderedParts:arePartsValid(attachableObject)
    local spec = self.spec_attachableCylinderedParts
    return #spec.attachableCylinderedParts > 0
        and attachableObject.spec_attachableCylinderedParts ~= nil and #attachableObject.spec_attachableCylinderedParts.attachableCylinderedParts > 0
end

---Connect movingPart to attachedVehicle
---@param part table
---@param partAttached table
---@param attachedVehicle table
---@return boolean
function AttachableCylinderedParts:connectMovingParts(part, partAttached, attachedVehicle)
    local retVal = false
    part.connectedPart = nil

    if part.node ~= nil then
        --get references
        if partAttached.referenceFrame ~= nil then
            --replace referenceFrame
            local movingPart = self:getMovingPartByNode(part.node)
            if movingPart ~= nil then
                movingPart.referenceFrameBackUp = movingPart.referenceFrame
                movingPart.referenceFrame = partAttached.referenceFrame
                retVal = true
                part.connectedPart = partAttached
            end
        elseif partAttached.referencePoint ~= nil then
            --replace referencePoint
            local movingPart = self:getMovingPartByNode(part.node)
            if movingPart ~= nil then
                movingPart.referencePointBackUp = movingPart.referencePoint
                movingPart.referencePoint = partAttached.referencePoint
                retVal = true
                part.connectedPart = partAttached
            end
        end
    elseif part.referenceFrame ~= nil then
        --replace referenceFrame
        if partAttached.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(partAttached.node)
            if movingPartAttached ~= nil then
                movingPartAttached.referenceFrameBackUp = movingPartAttached.referenceFrame
                movingPartAttached.referenceFrame = part.referenceFrame
                retVal = true
                part.connectedPart = partAttached
            end
        end
    elseif part.referencePoint ~= nil then
        --replace referencePoint        
        if partAttached.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(partAttached.node)
            if movingPartAttached ~= nil then
                movingPartAttached.referencePointBackUp = movingPartAttached.referencePoint
                movingPartAttached.referencePoint = part.referencePoint
                retVal = true
                part.connectedPart = partAttached
            end
        end
    end

    return retVal
end

---Disconnect movingPart from attachedVehicle
---@param part table
---@param attachedVehicle table
---@return boolean
function AttachableCylinderedParts:disconnectMovingParts(part, attachedVehicle)
    local retVal = false

    if part.node ~= nil then
        --get references
        if part.connectedPart.referenceFrame ~= nil then
            --replace referenceFrame
            local movingPart = self:getMovingPartByNode(part.node)
            if movingPart ~= nil then
                movingPart.referenceFrame = movingPart.referenceFrameBackUp
                retVal = true
            end
        elseif part.connectedPart.referencePoint ~= nil then
            --replace referencePoint
            local movingPart = self:getMovingPartByNode(part.node)
            if movingPart ~= nil then
                movingPart.referencePoint = movingPart.referencePointBackUp
                retVal = true
            end
        end
    elseif part.referenceFrame ~= nil then
        --replace referenceFrame
        if part.connectedPart.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(part.connectedPart.node)
            if movingPartAttached ~= nil then
                movingPartAttached.referenceFrame = movingPartAttached.referenceFrameBackUp
                retVal = true
            end
        end
    elseif part.referencePoint ~= nil then
        --replace referencePoint        
        if part.connectedPart.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(part.connectedPart.node)
            if movingPartAttached ~= nil then
                movingPartAttached.referencePoint = movingPartAttached.referencePointBackUp
                retVal = true
            end
        end
    end

    return retVal
end
