----------------------------------------------------------------------------------------------------
-- DynamicCultivatorMode
----------------------------------------------------------------------------------------------------
-- Purpose: Specialization for dynamic cultivator mode setting by animation
--
-- @author John Deere 6930 @VertexDezign
----------------------------------------------------------------------------------------------------

---@class DynamicCultivatorMode
DynamicCultivatorMode = {}
DynamicCultivatorMode.MOD_NAME = g_currentModName

function DynamicCultivatorMode.initSpecialization(specializations)
    local schema = Vehicle.xmlSchema
    schema:setXMLSpecializationType("Cultivator")
    schema:register(XMLValueType.BOOL, "vehicle.cultivator.dynamicMode#value", "Use dynamic cultivator mode", false)
    schema:register(XMLValueType.STRING, "vehicle.cultivator.dynamicMode#animationName", "Animation name to look at.")
    schema:register(XMLValueType.FLOAT, "vehicle.cultivator.dynamicMode#deepModeTime", "Animation time to use deepMode mode.")
    schema:register(XMLValueType.FLOAT, "vehicle.cultivator.dynamicMode#powerHarrowTime", "Animation time to use powerHarrow mode.")
    schema:register(XMLValueType.FLOAT, "vehicle.cultivator.dynamicMode#subsoilerTime", "Animation time to use subsoiler mode.")
    schema:setXMLSpecializationType()

    schema:setXMLSpecializationType("PowerConsumer")
    schema:register(XMLValueType.FLOAT, "vehicle.powerConsumer.dynamicCultivatorPowerMultiplier#startMultiplier", "Start multiplier if no dynamic depth is applied.")
    schema:register(XMLValueType.FLOAT, "vehicle.powerConsumer.dynamicCultivatorPowerMultiplier#endMultiplier", "End multiplier if all dynamic depth is applied.")
    schema:setXMLSpecializationType()
end

function DynamicCultivatorMode.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Cultivator, specializations)
end

function DynamicCultivatorMode.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "updateDynamicModes", DynamicCultivatorMode.updateDynamicModes)
end

function DynamicCultivatorMode.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", DynamicCultivatorMode)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", DynamicCultivatorMode)
	SpecializationUtil.registerEventListener(vehicleType, "onDraw", DynamicCultivatorMode)
end

function DynamicCultivatorMode.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getPowerMultiplier", DynamicCultivatorMode.getPowerMultiplier)
end

---Called on load
function DynamicCultivatorMode:onLoad(savegame)
    self.spec_dynamicCultivatorMode = self[string.format("spec_%s.dynamicCultivatorMode", DynamicCultivatorMode.MOD_NAME)]
    local spec = self.spec_dynamicCultivatorMode

	spec.useDynamicMode = self.xmlFile:getValue("vehicle.cultivator.dynamicMode#value", false)

    if spec.useDynamicMode then
        spec.animationName = self.xmlFile:getValue("vehicle.cultivator.dynamicMode#animationName")

        if spec.animationName == nil then
            Logging.xmlWarning(self.xmlFile, "No animationName found '%s'!", "vehicle.cultivator.dynamicMode#animationName")
            spec.useDynamicMode = false

            return
        end

        if not self:getAnimationExists(spec.animationName)then
            Logging.xmlWarning(self.xmlFile, "No valid animation found with name '%s'!", spec.animationName)
            spec.useDynamicMode = false

            return
        end

        spec.deepModeTime = self.xmlFile:getValue("vehicle.cultivator.dynamicMode#deepModeTime")
        spec.powerHarrowTime = self.xmlFile:getValue("vehicle.cultivator.dynamicMode#powerHarrowTime")
        spec.subsoilerTime = self.xmlFile:getValue("vehicle.cultivator.dynamicMode#subsoilerTime")

        spec.startMultiplier = self.xmlFile:getValue("vehicle.powerConsumer.dynamicCultivatorPowerMultiplier#startMultiplier")
        local endMultiplier = self.xmlFile:getValue("vehicle.powerConsumer.dynamicCultivatorPowerMultiplier#endMultiplier")

        if spec.startMultiplier ~= nil and endMultiplier ~= nil then
            spec.multiplierRange = endMultiplier - spec.startMultiplier
        end
    end

    spec.lastUpdatedAnimationTime = nil
    spec.animationTimeUpdateInterval = 0.05
end

---Called on update tick
function DynamicCultivatorMode:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_dynamicCultivatorMode

    if spec.useDynamicMode then
        local animTime = self:getAnimationTime(spec.animationName)
        if spec.lastUpdatedAnimationTime ~= nil then
            if math.abs(animTime - spec.lastUpdatedAnimationTime) >= spec.animationTimeUpdateInterval then
                self:updateDynamicModes(animTime)
            end
        else
            self:updateDynamicModes(animTime)
        end
    end
end

---Called on draw
function DynamicCultivatorMode:onDraw(isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_dynamicCultivatorMode

    if isActiveForInputIgnoreSelection and spec.lastUpdatedAnimationTime ~= nil then
        local activeText = g_i18n:getText("info_cultivatorMode_flat")

        if spec.deepModeTime ~= nil then
            if spec.lastUpdatedAnimationTime > spec.deepModeTime then
                activeText = g_i18n:getText("info_cultivatorMode_deep")
            end
        end
        if spec.powerHarrowTime ~= nil then
            if spec.lastUpdatedAnimationTime > spec.powerHarrowTime then
                activeText = g_i18n:getText("info_cultivatorMode_powerHarrow")
            end
        end
        if spec.subsoilerTime ~= nil then
            if spec.lastUpdatedAnimationTime > spec.subsoilerTime then
                activeText = g_i18n:getText("info_cultivatorMode_subsoiler")
            end
        end

        g_currentMission:addExtraPrintText(activeText)
        g_currentMission:addExtraPrintText(g_i18n:getText("info_cultivatorMode"))
    end
end

---Updates dynamicModes of cultivator
---@param animTime number
function DynamicCultivatorMode:updateDynamicModes(animTime)
    local spec = self.spec_dynamicCultivatorMode
    if not spec.useDynamicMode then
        return
    end

    local spec_cultivator = self.spec_cultivator
    local changed = false

    if spec.deepModeTime ~= nil then
        local value = animTime > spec.deepModeTime
        if value ~= spec_cultivator.useDeepMode then
            spec_cultivator.useDeepMode = value

            changed = true
        end
    end
    if spec.powerHarrowTime ~= nil then
        local value = animTime > spec.powerHarrowTime
        if value ~= spec_cultivator.isPowerHarrow then
            spec_cultivator.isPowerHarrow = value

            changed = true
        end
    end
    if spec.subsoilerTime ~= nil then
        local value = animTime > spec.subsoilerTime
        if value ~= spec_cultivator.isSubsoiler then
            spec_cultivator.isSubsoiler = value

            changed = true
        end
    end

    if changed then
        self:updateCultivatorAIRequirements()
        self:updateCultivatorEnabledState()
    end

    spec.lastUpdatedAnimationTime = animTime
end

---Overwritten function getPowerMultiplier
---Injects dynamic cultivator mode multiplication range
---@param superFunc function
---@return number
function DynamicCultivatorMode:getPowerMultiplier(superFunc)
    local spec = self.spec_dynamicCultivatorMode
    if not spec.useDynamicMode or spec.multiplierRange == nil then
        return superFunc(self)
    end

    return superFunc(self) * (spec.startMultiplier + spec.multiplierRange * spec.lastUpdatedAnimationTime)
end
