----------------------------------------------------------------------------------------------------
-- SupportAnimationAttachedUpdate
----------------------------------------------------------------------------------------------------
-- Purpose: Specialization for updating support animation on attached vehicles
--
-- @author John Deere 6930 @VertexDezign
----------------------------------------------------------------------------------------------------

---@class SupportAnimationAttachedUpdate
SupportAnimationAttachedUpdate = {}
SupportAnimationAttachedUpdate.MOD_NAME = g_currentModName

function SupportAnimationAttachedUpdate.initSpecialization(specializations)
    local schema = Vehicle.xmlSchema
    schema:setXMLSpecializationType("AttacherJoints")
    schema:register(XMLValueType.BOOL, "vehicle.attacherJoints.supportAnimation#updateAttached", "Updated attached vehicle supportAnimations", false)
    schema:register(XMLValueType.BOOL, "vehicle.attacherJoints.supportAnimation#invertAttached", "Invert attached vehicle supportAnimations", false)
    schema:setXMLSpecializationType()
end

function SupportAnimationAttachedUpdate.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(AttacherJoints, specializations)
end

function SupportAnimationAttachedUpdate.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "updateAttachedSupportAnimations", SupportAnimationAttachedUpdate.updateAttachedSupportAnimations)
end

function SupportAnimationAttachedUpdate.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", SupportAnimationAttachedUpdate)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", SupportAnimationAttachedUpdate)
    SpecializationUtil.registerEventListener(vehicleType, "onPreAttach", SupportAnimationAttachedUpdate)
    SpecializationUtil.registerEventListener(vehicleType, "onPostAttach", SupportAnimationAttachedUpdate)
    SpecializationUtil.registerEventListener(vehicleType, "onPostDetach", SupportAnimationAttachedUpdate)
end

---Called on load
function SupportAnimationAttachedUpdate:onLoad(savegame)
    self.spec_supportAnimationAttachedUpdate = self[string.format("spec_%s.supportAnimationAttachedUpdate", SupportAnimationAttachedUpdate.MOD_NAME)]
    local spec = self.spec_supportAnimationAttachedUpdate

    spec.updateAttached = self.xmlFile:getValue("vehicle.attacherJoints.supportAnimation#updateAttached", false)
    spec.invertAttached = self.xmlFile:getValue("vehicle.attacherJoints.supportAnimation#invertAttached", false)

    spec.initAttachedVehicles = false
    spec.initAttachedVehiclesTime = g_currentMission.time + 2000
end

---Called on update
function SupportAnimationAttachedUpdate:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_supportAnimationAttachedUpdate

    if not spec.initAttachedVehicles and spec.initAttachedVehiclesTime < g_currentMission.time then
        if self:getAttacherVehicle() == nil then
            self:updateAttachedSupportAnimations(1)
        end
        spec.initAttachedVehicles = true
    end
end

---Called before attach
function SupportAnimationAttachedUpdate:onPreAttach(attacherVehicle, inputJointDescIndex, jointDescIndex)
    self:updateAttachedSupportAnimations(-1)
end

---Called after attach
function SupportAnimationAttachedUpdate:onPostAttach(attacherVehicle, inputJointDescIndex, jointDescIndex)
    self:updateAttachedSupportAnimations(-1)
end

---Called after detach
function SupportAnimationAttachedUpdate:onPostDetach()
    self:updateAttachedSupportAnimations(1)
end

---Updates the supportAnimations of attached implements
---@param direction number
function SupportAnimationAttachedUpdate:updateAttachedSupportAnimations(direction)
    local spec = self.spec_supportAnimationAttachedUpdate

    if spec.updateAttached then
        direction = spec.invertAttached and -direction or direction

        for _, implement in ipairs(self.spec_attacherJoints.attachedImplements) do
            if implement.object ~= nil and implement.object.spec_attachable ~= nil then
                for _, supportAnimation in ipairs(implement.object.spec_attachable.supportAnimations) do
                    if implement.object:getIsSupportAnimationAllowed(supportAnimation) then
                        implement.object:playAnimation(supportAnimation.animationName, direction, nil, true)

                        if implement.object.propertyState == Vehicle.PROPERTY_STATE_SHOP_CONFIG then
                            AnimatedVehicle.updateAnimationByName(implement.object, supportAnimation.animationName, 9999999)
                        end
                    end
                end
            end
        end
    end
end
